
from __future__ import annotations

__all__ = ('Config',)

import os
from pathlib import Path
from typing import Union

import click

from .exceptions import ConfigError


_MaybePath = Union[None, Path, str]


def _normalise_path(path: Union[Path, str]) -> Path:
    return Path(path).expanduser().resolve()


class Config:

    config_dir: Path
    plugin_dir: Path
    addon_dir:  Path

    def __init__(self, *, config_dir: _MaybePath = None,
                 addon_dir: _MaybePath = None) -> None:
        def _config_dir():
            yield os.environ.get('INSTAWOW_CONFIG_DIR')
            yield config_dir
            yield click.get_app_dir('instawow')

        def _addon_dir():
            yield addon_dir
            yield ((self.config_dir / 'addon_dir.txt')
                   .read_text(encoding='utf-8').strip())

        self.config_dir = _normalise_path(next(p for p in _config_dir() if p))
        self.plugin_dir = self.config_dir / 'plugins'

        try:
            self.addon_dir = _normalise_path(next(p for p in _addon_dir() if p))
        except FileNotFoundError:
            raise ConfigError('configuration not written on disk')
        if not self.addon_dir.is_dir():
            raise ConfigError(f"'{self.addon_dir}' is not a directory")

    def write(self) -> Config:
        self.config_dir.mkdir(exist_ok=True)
        self.plugin_dir.mkdir(exist_ok=True)
        (self.config_dir / 'addon_dir.txt').write_text(str(self.addon_dir),
                                                       encoding='utf-8')
        return self




from __future__ import annotations

__all__ = ('Strategies', 'Pkg_',
           'CurseResolver', 'WowiResolver', 'TukuiResolver', 'InstawowResolver')

import asyncio
from datetime import datetime
from enum import Enum
from typing import TYPE_CHECKING, Any, Callable, ClassVar, Optional, Set, Tuple

from yarl import URL

from . import exceptions as E
from .models import Pkg, PkgCoercer, PkgOptions
from .utils import ManagerAttrAccessMixin, slugify

if TYPE_CHECKING:
    from .manager import Manager


class Strategies(str, Enum):

    default = 'default'
    latest = 'latest'

    @classmethod
    def validate(cls, method: Callable) -> Callable:
        def wrapper(self, id_or_slug: str, *, strategy: str) -> Callable:
            try:
                strategy_enum = cls[strategy]
            except KeyError:
                pass
            else:
                if strategy_enum in self.strategies:
                    return method(self, id_or_slug, strategy=strategy_enum)
            raise E.PkgStrategyInvalid(strategy)

        return wrapper


class Pkg_(PkgCoercer):

    def __call__(self, **kwargs: Any) -> Pkg:
        return Pkg(**self.dict(), **kwargs)


class Resolver(ManagerAttrAccessMixin):

    origin: ClassVar[str]
    name: ClassVar[str]
    strategies: ClassVar[Set[Strategies]]

    def __init__(self, *, manager: Manager) -> None:
        self.manager = manager

    @classmethod
    def decompose_url(cls, uri: str) -> Optional[Tuple[str, str]]:
        "Break a URL down to its component `origin` and `id`."
        raise NotImplementedError

    async def resolve(self, id_or_slug: str, *, strategy: Strategies) -> Pkg_:
        "Turn an ID or slug into a `Pkg_`."
        raise NotImplementedError


class CurseResolver(Resolver):

    origin = 'curse'
    name = 'CurseForge'
    strategies = {Strategies.default, Strategies.latest}

    curse_url = URL('https://www.curseforge.com/wow/addons')
    # https://twitchappapi.docs.apiary.io/
    addon_api_url = URL('https://addons-ecs.forgesvc.net/api/v2/addon')

    @classmethod
    def decompose_url(cls, uri: str) -> Optional[Tuple[str, str]]:
        url = URL(uri)
        if url.host == 'www.wowace.com' \
                and len(url.parts) > 2 \
                and url.parts[1] == 'projects':
            return (cls.origin, url.parts[2])
        elif url.host == 'www.curseforge.com' \
                and len(url.parts) > 3 \
                and url.parts[1:3] == ('wow', 'addons'):
            return (cls.origin, url.parts[3])

    @Strategies.validate
    async def resolve(self, id_or_slug: str, *, strategy: Strategies) -> Pkg_:
        from lxml.html import document_fromstring

        project_url = self.curse_url / id_or_slug
        async with self.web_client.get(project_url) as response:
            if response.status == 404:
                addon_id = id_or_slug
            else:
                html = document_fromstring(await response.text())
                addon_id, = html.xpath('//span[text() = "Project ID"]'
                                       '/following-sibling::span'
                                       '/text()')

        url = self.addon_api_url / addon_id
        async with self.web_client.get(url) as response:
            if response.status == 404:
                raise E.PkgNonexistent
            addon_metadata = await response.json()

        if strategy is Strategies.default:
            file = next(f for f in addon_metadata['latestFiles']
                        if f['id'] == addon_metadata['defaultFileId'])
        elif strategy is Strategies.latest:
            _, file = max((f['id'], f) for f in addon_metadata['latestFiles']
                          if not f['isAlternate'])

        return Pkg_(origin=self.origin,
                    id=addon_metadata['id'],
                    slug=addon_metadata['slug'],
                    name=addon_metadata['name'],
                    description=addon_metadata['summary'],
                    url=addon_metadata['websiteUrl'],
                    file_id=file['id'],
                    download_url=file['downloadUrl'],
                    date_published=file['fileDate'],
                    version=file['displayName'],
                    options=PkgOptions(strategy=strategy.name))


class WowiResolver(Resolver):

    origin = 'wowi'
    name = 'WoWInterface'
    strategies = {Strategies.default}

    # https://api.mmoui.com/v3/globalconfig.json
    list_api_url = URL('https://api.mmoui.com/v3/game/WOW/filelist.json')
    details_api_url = URL('https://api.mmoui.com/v3/game/WOW/filedetails/')

    def __init__(self, *, manager: Manager) -> None:
        super().__init__(manager=manager)
        self.files: Any = None
        self._sync_lock = asyncio.Lock()

    async def _sync(self) -> None:
        async with self._sync_lock:
            if not self.files:
                async with self.web_client.get(self.list_api_url) as response:
                    self.files = {i['UID']: i for i in await response.json()}

                async def noop() -> None:
                    pass
                setattr(self, '_sync', noop)

    @classmethod
    def decompose_url(cls, uri: str) -> Optional[Tuple[str, str]]:
        url = URL(uri)
        if url.host in {'wowinterface.com', 'www.wowinterface.com'} \
                and len(url.parts) == 3 \
                and url.parts[1] == 'downloads':
            if url.name == 'landing.php':
                id_ = url.query.get('fileid')
                if id_:
                    return (cls.origin, id_)
            elif url.name.startswith('info'):
                from itertools import takewhile
                id_ = ''.join(takewhile(str.isdigit, url.name[4:]))
                if id_:
                    return (cls.origin, id_)

    @Strategies.validate
    async def resolve(self, id_or_slug: str, *, strategy: Strategies) -> Pkg_:
        await self._sync()
        try:
            addon = self.files[id_or_slug.partition('-')[0]]
        except KeyError:
            raise E.PkgNonexistent

        url = self.details_api_url / f'{addon["UID"]}.json'
        async with self.web_client.get(url) as response:
            addon_details, = await response.json()
        if addon_details['UIPending'] == '1':
            raise E.PkgTemporarilyUnavailable

        return Pkg_(origin=self.origin,
                    id=addon['UID'],
                    slug=slugify(f'{addon["UID"]} {addon["UIName"]}'),
                    name=addon['UIName'],
                    description=addon_details['UIDescription'],
                    url=addon['UIFileInfoURL'],
                    file_id=addon_details['UIMD5'],
                    download_url=addon_details['UIDownload'],
                    date_published=addon_details['UIDate'],
                    version=addon_details['UIVersion'],
                    options=PkgOptions(strategy=strategy.name))


class TukuiResolver(Resolver):

    origin = 'tukui'
    name = 'Tukui'
    strategies = {Strategies.default}

    api_url = URL('https://www.tukui.org/api.php')

    @classmethod
    def decompose_url(cls, uri: str) -> Optional[Tuple[str, str]]:
        url = URL(uri)
        if url.host == 'www.tukui.org' \
                and url.path in {'/addons.php', '/download.php'}:
            id_or_slug = url.query.get('id') or url.query.get('ui')
            if id_or_slug:
                return (cls.origin, id_or_slug)

    @Strategies.validate
    async def resolve(self, id_or_slug: str, *, strategy: Strategies) -> Pkg_:
        id_or_slug = id_or_slug.partition('-')[0]
        is_ui = id_or_slug in {'elvui', 'tukui'}

        url = self.api_url.with_query({'ui' if is_ui else 'addon': id_or_slug})
        async with self.web_client.get(url) as response:
            if not response.content_length:
                raise E.PkgNonexistent
            addon = await response.json(content_type='text/html')

        return Pkg_(origin=self.origin,
                    id=addon['id'],
                    slug=id_or_slug if is_ui else slugify(f'{addon["id"]} {addon["name"]}'),
                    name=addon['name'],
                    description=addon['small_desc'],
                    url=addon['web_url'],
                    file_id=addon['lastupdate'],
                    download_url=addon['url'],
                    date_published=(datetime.fromisoformat(addon['lastupdate'])
                                    if is_ui else addon['lastupdate']),
                    version=addon['version'],
                    options=PkgOptions(strategy=strategy.name))


class InstawowResolver(Resolver):

    origin = 'instawow'
    name = 'instawow'
    strategies = {Strategies.default}

    @classmethod
    def decompose_url(cls, uri: str) -> None:
        return

    @Strategies.validate
    async def resolve(self, id_or_slug: str, *, strategy: Strategies) -> Pkg_:
        if id_or_slug not in {'0', 'weakauras-companion'}:
            raise E.PkgNonexistent

        from .manager import run_in_thread
        from .wa_updater import WaCompanionBuilder

        builder = WaCompanionBuilder(self.manager)

        return Pkg_(origin=self.origin,
                    id='0',
                    slug='weakauras-companion',
                    name='WeakAuras Companion',
                    description='A WeakAuras Companion wannabe.',
                    url='https://github.com/layday/instawow',
                    file_id=await run_in_thread(builder.checksum)(),
                    download_url=builder.file_out.as_uri(),
                    date_published=datetime.now(),
                    version='1.0.0',
                    options=PkgOptions(strategy=strategy.name))




from __future__ import annotations

__all__ = ('WaCompanionBuilder',)

import asyncio
from functools import partial, reduce
from pathlib import Path
from typing import TYPE_CHECKING
from typing import Any, Callable, Dict, Iterable, List, Optional, Sequence, Tuple, Union

from loguru import logger
from pydantic import BaseModel, Extra, validator
from yarl import URL

from .utils import ManagerAttrAccessMixin, bucketise

if TYPE_CHECKING:
    from luaparser import astnodes as lua_nodes
    from .manager import Manager


metadata_api = URL('https://data.wago.io/api/check/weakauras')
raw_api = URL('https://data.wago.io/api/raw/encoded')

_EMPTY_URL = URL()


class AuraURLInvalid(Exception):
    pass


class AuraEntry(BaseModel):

    id: str
    uid: str
    parent: Optional[str]
    url: URL = _EMPTY_URL
    version: int
    semver: Optional[str]
    ignore_wago_update: bool = False

    class Config:
        arbitrary_types_allowed = True
        fields = {'ignore_wago_update' : {'alias': 'ignoreWagoUpdate'}}

    @validator('url', always=True, pre=True)
    def __prep_url(cls, value: Any) -> Any:
        if value == _EMPTY_URL:
            raise AuraURLInvalid
        return value

    @validator('id', 'uid', 'parent', 'url', 'semver', pre=True)
    def __convert_str(cls, value: lua_nodes.String) -> str:
        return value.s

    @validator('version', pre=True)
    def __convert_int(cls, value: lua_nodes.Number) -> int:
        return value.n

    @validator('ignore_wago_update', pre=True)
    def __convert_bool(cls, value: Union[lua_nodes.TrueExpr, lua_nodes.FalseExpr]
                       ) -> bool:
        return value.display_name == 'True'

    @validator('url', pre=True)
    def __postp_url(cls, value: str) -> URL:
        url = URL(value)
        if url.host != 'wago.io':
            logger.info(f'discarding aura with URL: {url}')
            raise AuraURLInvalid
        return url

    @classmethod
    def from_lua_ast(cls, tree: lua_nodes.Field) -> Optional[AuraEntry]:
        try:
            return cls(**{f.key.s: f.value for f in tree.value.fields})
        # The URL validators raise an exception not derived from ValueError
        # so that they won't be wrapped by Pydantic and we can catch them here
        except AuraURLInvalid:
            return None


class ApiMetadata__Changelog(BaseModel):

    format: Optional[str]
    text: str = ''

    class Config:
        extra = Extra.forbid


class ApiMetadata(BaseModel):

    id: str
    name: str
    slug: str
    url: str
    created: str
    modified: str
    game: str
    fork_of: Optional[str]
    username: str
    version: int
    version_string: str
    changelog: ApiMetadata__Changelog
    region_type: Optional[str]

    class Config:
        extra = Extra.forbid
        fields = {'id': {'alias': '_id'},
                  'fork_of': {'alias': 'forkOf'},
                  'version_string': {'alias': 'versionString'},
                  'region_type': {'alias': 'regionType'}}


_T_Installed = Dict[str, List[AuraEntry]]
_T_Outdated = List[Tuple[str, List[AuraEntry], ApiMetadata, str]]


class WaCompanionBuilder(ManagerAttrAccessMixin):
    """A WeakAuras Companion port for shellfolk."""

    def __init__(self, manager: Manager) -> None:
        self.manager = manager
        self.builder_dir = self.manager.config.plugin_dir / __name__
        self.file_out = self.builder_dir / 'WeakAurasCompanion.zip'

    @staticmethod
    def extract_auras(source: str) -> _T_Installed:
        from luaparser import ast as lua_ast

        lua_tree = lua_ast.walk(lua_ast.parse(source))
        aura_table = next(i for i in lua_tree
                          if reduce(lambda p, n: getattr(p, n, None),
                                    [i, 'key', 's']) == 'displays').value.fields
        auras = filter(None, map(AuraEntry.from_lua_ast, aura_table))
        aura_groups = bucketise(auras, key=lambda a: a.url.parts[1])
        aura_groups = {k: v for k, v in aura_groups.items()
                       if not any(i for i in v if i.ignore_wago_update)}
        return aura_groups

    def extract_installed_auras(self, account: str) -> _T_Installed:
        import time

        accounts = self.config.addon_dir.parents[1] / 'WTF/Account'
        saved_vars = accounts / account / 'SavedVariables/WeakAuras.lua'

        start = time.perf_counter()
        aura_groups = self.extract_auras(saved_vars.read_text(encoding='utf-8'))
        logger.debug(f'auras extracted in {time.perf_counter() - start}s')
        return aura_groups

    async def get_wago_aura_metadata(self, aura_ids: Sequence[str]) -> List[ApiMetadata]:
        url = metadata_api.with_query({'ids': ','.join(aura_ids)})
        async with self.web_client.get(url) as response:
            metadata = await response.json()

        results = dict.fromkeys(aura_ids)
        for item in (ApiMetadata(**i) for i in metadata):
            if item.slug in results:
                results[item.slug] = item
        return list(results.values())

    async def get_wago_aura(self, aura_id: str) -> str:
        url = raw_api.with_query({'id': aura_id})
        async with self.web_client.get(url) as response:
            return await response.text()

    async def get_outdated(self, aura_groups: _T_Installed) -> _T_Outdated:
        if not aura_groups:
            return []

        aura_metadata = await self.get_wago_aura_metadata(list(aura_groups))
        auras_with_metadata = filter(lambda v: v[1],
                                     zip(aura_groups.items(), aura_metadata))
        outdated_auras = [((s, w), r) for (s, w), r in auras_with_metadata
                          if r.version > next((a for a in w if not a.parent), w[0]).version
                          ]
        if not outdated_auras:
            return []

        new_auras = await asyncio.gather(*(self.get_wago_aura(r.id)
                                           for _, r in outdated_auras))
        return [(s, w, r, p)
                for ((s, w), r), p in zip(outdated_auras, new_auras)]

    def make_addon(self, outdated: _T_Outdated) -> None:
        from jinja2 import Environment, FileSystemLoader
        from zipfile import ZipFile, ZipInfo

        tpl_dir = Path(__file__).parent / 'wa_templates'
        jinja_env = Environment(loader=FileSystemLoader(str(tpl_dir)),
                                trim_blocks=True, lstrip_blocks=True)

        with ZipFile(self.file_out, 'w') as addon_zip:
            def write_tpl(filename: str, ctx: dict) -> None:
                # Not using a plain string as the first argument to ``writestr``
                # 'cause the timestamp is set dynamically by default, which
                # renders the build rather - shall we say - unreproducible
                zip_info = ZipInfo(filename=f'WeakAurasCompanion/{filename}')
                tpl = jinja_env.get_template(filename)
                addon_zip.writestr(zip_info, tpl.render(**ctx))

            write_tpl('data.lua',
                      {'was': [(m.slug,
                                {'name': m.name,
                                 'author': m.username,
                                 'encoded': p,
                                 'wagoVersion': m.version,
                                 'wagoSemver': m.version_string,
                                 'versionNote': m.changelog.text})
                               for _, _, m, p, in outdated],
                       'uids': [(a.uid, a.url.parts[1])
                                for _, w, _, _ in outdated
                                for a in w],
                       'ids':  [(a.id, a.url.parts[1])
                                for _, w, _, _ in outdated
                                for a in w],
                       'stash': []})    # Send to WAC not supported - always empty
            write_tpl('init.lua', {})
            write_tpl('WeakAurasCompanion.toc', {})

    def build(self, account: str) -> None:
        self.builder_dir.mkdir(exist_ok=True)
        installed = self.extract_installed_auras(account)
        outdated = self.manager.run(self.get_outdated(installed))
        self.make_addon(outdated)

    def checksum(self) -> str:
        from hashlib import sha256
        return sha256(self.file_out.read_bytes()).hexdigest()




from __future__ import annotations

__all__ = ('ManagerAttrAccessMixin',
           'TocReader',
           'bucketise',
           'slugify',
           'is_outdated',
           'setup_logging')

from collections import defaultdict, namedtuple
from datetime import datetime
from pathlib import Path
import re
from typing import TYPE_CHECKING
from typing import Any, Callable, Iterable, Optional, List, Tuple, Type, TypeVar, Union

from . import __version__

if TYPE_CHECKING:
    from .config import Config
    from .manager import Manager


O = TypeVar('O')


class cached_property:

    def __init__(self, func: Callable) -> None:
        self.func = func

    def __get__(self, obj: O, class_: Optional[Type[O]] = None) -> Any:
        if class_ is None:
            return self

        value = self.func(obj)
        obj.__dict__[self.func.__name__] = value
        return value


class ManagerAttrAccessMixin:

    def __getattr__(self, name: str) -> Any:
        return getattr(self.manager, name)


class TocReader:
    """Extracts key–value pairs from TOC files."""

    Entry = namedtuple('_TocEntry', 'key value')

    def __init__(self, path: Path, default: Union[None, str] = '') -> None:
        entries = (e.lstrip('# ').partition(': ')[::2]
                   for e in path.read_text(encoding='utf-8-sig').splitlines()
                   if e.startswith('## '))
        self.entries = dict(entries)
        self.default = default

    def __getitem__(self, key: Union[str, Tuple[str, ...]]) -> Entry:
        if isinstance(key, tuple):
            try:
                return next(filter(lambda i: i.value,
                                   (self.__getitem__(k) for k in key)))
            except StopIteration:
                key = key[0]
        return self.Entry(key, self.entries.get(key, self.default))


def bucketise(iterable: Iterable, key: Callable = (lambda v: v)) -> dict:
    "Place the elements of `iterable` into a bucket according to `key`."
    bucket = defaultdict(list)      # type: ignore
    for value in iterable:
        bucket[key(value)].append(value)
    return dict(bucket)


_match_loweralphanum = re.compile(r'[^0-9a-z ]')

def slugify(text: str) -> str:
    "Convert an add-on name into a lower-alphanumeric slug."
    return '-'.join(_match_loweralphanum.sub(' ', text.casefold()).split())


def is_outdated(manager: Manager) -> bool:
    """Check against PyPI to see if `instawow` is outdated.

    The response is cached for 24 hours.
    """
    def parse_version(version: str) -> Tuple[int, ...]:
        return tuple(map(int, version.split('.')))

    cache_file = manager.config.config_dir / '.pypi_version'
    mtime = cache_file.exists() and cache_file.stat().st_mtime
    if mtime and (datetime.now() - datetime.fromtimestamp(mtime)).days < 1:
        version = cache_file.read_text(encoding='utf-8')
    else:
        from aiohttp.client import ClientError

        async def get_metadata() -> dict:
            url = 'https://pypi.org/pypi/instawow/json'
            async with manager.web_client.get(url) as response:
                return await response.json()

        try:
            version = manager.run(get_metadata())['info']['version']
        except ClientError:
            version = __version__
        else:
            cache_file.write_text(version, encoding='utf-8')

    # Assume cache is stale if installed version > version on PyPI
    if parse_version(__version__) > parse_version(version):
        return False
    else:
        return __version__ != version


def setup_logging(config: Config, level: Union[int, str] = 'INFO') -> int:
    from loguru import logger

    handler = {'sink': config.config_dir / 'error.log',
               'level': level,
               'rotation': '1 MB',
               'enqueue': True}
    handler_id, = logger.configure(handlers=(handler,))
    return handler_id




from __future__ import annotations

__all__ = ('ModelBase',
           'Pkg',
           'PkgFolder',
           'PkgOptions',
           'PkgCoercer',
           'PkgFolderCoercer',
           'PkgOptionsCoercer',
           'should_migrate')

from datetime import datetime
from typing import TYPE_CHECKING

import pydantic
from sqlalchemy import Column, ForeignKeyConstraint, DateTime, String
import sqlalchemy.exc
from sqlalchemy.ext.declarative import declarative_base
from sqlalchemy.orm import relationship

if TYPE_CHECKING:
    import sqlalchemy.base


def _declarative_constructor(self, **kwargs):
    intermediate_obj = _COERCERS[self.__class__].parse_obj(kwargs)
    for k, v in intermediate_obj:
        setattr(self, k, v)


_declarative_constructor.__name__ = '__init__'

ModelBase = declarative_base(constructor=_declarative_constructor)


class _BaseCoercer(pydantic.BaseModel):
    """The coercer is used inside the declarative constructor to type-cast
    values _prior to_ inserts.  SQLAlchemy delegates this function to
    the DB API -- see https://stackoverflow.com/a/8980982.  We need this to
    happen in advance to be able to compare values with their in-database
    counterparts.  It also saves us the trouble of having to manually
    parse dates and the like.
    """

    class Config:
        extra = pydantic.Extra.allow
        max_anystr_length = 2 ** 32


class Pkg(ModelBase):

    __tablename__ = 'pkg'

    origin = Column(String, primary_key=True)
    id = Column(String, primary_key=True)
    slug = Column(String, nullable=False)
    name = Column(String, nullable=False)
    description = Column(String, nullable=False)
    url = Column(String, nullable=False)
    file_id = Column(String, nullable=False)
    download_url = Column(String, nullable=False)
    date_published = Column(DateTime, nullable=False)
    version = Column(String, nullable=False)
    folders = relationship('PkgFolder', cascade='all, delete-orphan',
                           backref='pkg')
    options = relationship('PkgOptions', cascade='all, delete-orphan',
                           uselist=False)


class PkgCoercer(_BaseCoercer):

    origin: str
    id: str
    slug: str
    name: str
    description: str
    url: str
    file_id: str
    download_url: str
    date_published: datetime
    version: str


class PkgFolder(ModelBase):

    __tablename__ = 'pkg_folder'
    __table_args__ = (ForeignKeyConstraint(['pkg_origin', 'pkg_id'],
                                           ['pkg.origin', 'pkg.id']),)

    name = Column(String, primary_key=True)
    pkg_origin = Column(String, nullable=False)
    pkg_id = Column(String, nullable=False)


class PkgFolderCoercer(_BaseCoercer):

    name: str


class PkgOptions(ModelBase):

    __tablename__ = 'pkg_options'
    __table_args__ = (ForeignKeyConstraint(['pkg_origin', 'pkg_id'],
                                           ['pkg.origin', 'pkg.id']),)

    strategy = Column(String, nullable=False)
    pkg_origin = Column(String, primary_key=True)
    pkg_id = Column(String, primary_key=True)


class PkgOptionsCoercer(_BaseCoercer):

    strategy: str


_COERCERS = {Pkg: PkgCoercer,
             PkgFolder: PkgFolderCoercer,
             PkgOptions: PkgOptionsCoercer}


def should_migrate(engine: sqlalchemy.base.Engine, version: str) -> bool:
    """Check if the database version is the same as `version`;
    if not a migration would be required.
    """
    with engine.begin() as conn:
        try:
            current = conn.execute('SELECT version_num '
                                   'FROM alembic_version '
                                   'WHERE version_num = (?)', version).scalar()
        except sqlalchemy.exc.OperationalError:
            return True
        else:
            if not current:
                return True
    return False




from click.testing import CliRunner
import pytest

from instawow.cli import main
from instawow.config import Config
from instawow.manager import CliManager


class _CliTest:

    @pytest.fixture(scope='class')
    def manager(self, tmp_path_factory, request):
        tmp_path = tmp_path_factory.mktemp('.'.join((__name__,
                                                     request.cls.__name__)))
        addons = tmp_path / 'addons'
        addons.mkdir()
        config = Config(config_dir=tmp_path / 'config', addon_dir=addons)
        config.write()
        yield CliManager(config, show_progress=False)

    @pytest.fixture(scope='class')
    def invoke_runner(self, manager):
        yield lambda args: CliRunner().invoke(main, args=args, obj=manager)


class TestValidCursePkgLifecycle(_CliTest):

    @pytest.mark.parametrize(
        'test_input, cmp_fn, expected_output',
        [('install curse:molinari',
          str.startswith,
          '✓ curse:molinari\n  installed'),
         ('install curse:molinari',
          str.__eq__,
          '✗ curse:molinari\n  package already installed\n'),
         ('update curse:molinari',
          str.__eq__,
          '✗ curse:molinari\n  package is up to date\n'),
         ('remove curse:molinari',
          str.__eq__,
          '✓ curse:molinari\n  removed\n'),
         ('update curse:molinari',
          str.__eq__,
          '✗ curse:molinari\n  package is not installed\n'),
         ('remove curse:molinari',
          str.__eq__,
          '✗ curse:molinari\n  package is not installed\n'),])
    def test_valid_curse_pkg_lifecycle(self, invoke_runner,
                                       test_input, cmp_fn, expected_output):
        assert cmp_fn(invoke_runner(test_input).output, expected_output)


class TestValidTukuiPkgLifecycle(_CliTest):

    @pytest.mark.parametrize(
        'test_input, cmp_fn, expected_output',
        [('install tukui:3',
          str.startswith,
          '✓ tukui:3\n  installed'),
         ('install tukui:3',
          str.__eq__,
          '✗ tukui:3\n  package already installed\n'),
         ('update tukui:3',
          str.__eq__,
          '✗ tukui:3\n  package is up to date\n'),
         ('remove tukui:3',
          str.__eq__,
          '✓ tukui:3\n  removed\n'),
         ('update tukui:3',
          str.__eq__,
          '✗ tukui:3\n  package is not installed\n'),
         ('remove tukui:3',
          str.__eq__,
          '✗ tukui:3\n  package is not installed\n'),
         ('install tukui:tukui',
          str.startswith,
          '✓ tukui:tukui\n  installed'),
         ('install tukui:tukui',
          str.__eq__,
          '✗ tukui:tukui\n  package already installed\n'),
         ('update tukui:tukui',
          str.__eq__,
          '✗ tukui:tukui\n  package is up to date\n'),
         ('remove tukui:tukui',
          str.__eq__,
          '✓ tukui:tukui\n  removed\n'),
         ('update tukui:tukui',
          str.__eq__,
          '✗ tukui:tukui\n  package is not installed\n'),
         ('remove tukui:tukui',
          str.__eq__,
          '✗ tukui:tukui\n  package is not installed\n'),
        ])
    @pytest.mark.skip(reason='corrupted zip files')
    def test_valid_tukui_pkg_lifecycle(self, invoke_runner,
                                       test_input, cmp_fn, expected_output):
        assert cmp_fn(invoke_runner(test_input).output, expected_output)


class TestValidWowiPkgLifecycle(_CliTest):

    @pytest.mark.parametrize(
        'test_input, cmp_fn, expected_output',
        [('install wowi:13188-molinari',
          str.startswith,
          '✓ wowi:13188-molinari\n  installed'),
         ('install wowi:13188-molinari',
          str.__eq__,
          '✗ wowi:13188-molinari\n  package already installed\n'),
         ('update wowi:13188-molinari',
          str.__eq__,
          '✗ wowi:13188-molinari\n  package is up to date\n'),
         ('remove wowi:13188-molinari',
          str.__eq__,
          '✓ wowi:13188-molinari\n  removed\n'),
         ('update wowi:13188-molinari',
          str.__eq__,
          '✗ wowi:13188-molinari\n  package is not installed\n'),
         ('remove wowi:13188-molinari',
          str.__eq__,
          '✗ wowi:13188-molinari\n  package is not installed\n'),
        ])
    def test_valid_wowi_pkg_lifecycle(self, invoke_runner,
                                      test_input, cmp_fn, expected_output):
        assert cmp_fn(invoke_runner(test_input).output, expected_output)


class TestFolderConflictLifecycle(_CliTest):

    @pytest.mark.parametrize('test_input, cmp_fn, expected_output',
                             [('install curse:molinari',
                               str.startswith,
                               '✓ curse:molinari\n  installed'),
                              ('install wowi:13188-molinari',
                               str.__eq__,
                               "✗ wowi:13188-molinari\n"
                               "  package folders conflict with installed package's curse:molinari\n"),
                              ('remove curse:molinari',
                               str.__eq__,
                               '✓ curse:molinari\n  removed\n'),
                             ])
    def test_folder_conflict_lifecycle(self, invoke_runner,
                                       test_input, cmp_fn, expected_output):
        assert cmp_fn(invoke_runner(test_input).output, expected_output)


class TestPreexistingFolderConflictOnInstall(_CliTest):

    def test_preexisting_folder_conflict_on_install(self, manager, invoke_runner):
        (manager.config.addon_dir / 'Molinari').mkdir()
        assert invoke_runner('install curse:molinari').output == \
            "✗ curse:molinari\n  package folders conflict with an add-on's not controlled by instawow\n"


class TestInvalidAddonNameLifecycle(_CliTest):

    @pytest.mark.parametrize('test_input, expected_output',
                             [('install curse:gargantuan-wigs',
                               '✗ curse:gargantuan-wigs\n  package does not exist\n'),
                              ('update curse:gargantuan-wigs',
                               '✗ curse:gargantuan-wigs\n  package is not installed\n'),
                              ('remove curse:gargantuan-wigs',
                               '✗ curse:gargantuan-wigs\n  package is not installed\n')
                             ])
    def test_invalid_addon_name_lifecycle(self, invoke_runner,
                                          test_input, expected_output):
        assert invoke_runner(test_input).output == expected_output


class TestInvalidOriginLifecycle(_CliTest):

    @pytest.mark.parametrize('test_input, expected_output',
                             [('install foo:bar',
                               '✗ foo:bar\n  package origin is invalid\n'),
                              ('update foo:bar',
                               '✗ foo:bar\n  package is not installed\n'),
                              ('remove foo:bar',
                               '✗ foo:bar\n  package is not installed\n')
                             ])
    def test_invalid_origin_lifecycle(self, invoke_runner,
                                      test_input, expected_output):
        assert invoke_runner(test_input).output == expected_output


class TestStrategySwitchAndUpdateLifecycle(_CliTest):

    @pytest.mark.parametrize('test_input, cmp_fn, expected_output',
                             [('install curse:kuinameplates',
                               str.startswith,
                               '✓ curse:kuinameplates\n  installed'),
                              ('update --strategy=latest curse:kuinameplates',
                               str.startswith,
                               '✓ curse:kuinameplates\n  updated'),
                              ('update --strategy=default curse:kuinameplates',
                               str.startswith,
                               '✓ curse:kuinameplates\n  updated'),
                              ('remove curse:kuinameplates',
                               str.__eq__,
                               '✓ curse:kuinameplates\n  removed\n'),
                             ])
    def test_strategy_switch_and_update_lifecycle(self, invoke_runner,
                                                  test_input, cmp_fn,
                                                  expected_output):
        assert cmp_fn(invoke_runner(test_input).output, expected_output)


class TestInstallWithAlias(_CliTest):

    @pytest.mark.parametrize(
        'test_input, cmp_fn, expected_output',
        [('install curse:molinari',
          str.startswith,
          '✓ curse:molinari\n  installed'),
         ('install curse:20338',
          str.__eq__,
          '✗ curse:20338\n  package already installed\n'),
         ('install https://www.wowace.com/projects/molinari',
          str.__eq__,
          '✗ curse:molinari\n  package already installed\n'),
         ('install https://www.curseforge.com/wow/addons/molinari',
          str.__eq__,
          '✗ curse:molinari\n  package already installed\n'),
         ('install https://www.curseforge.com/wow/addons/molinari/download',
          str.__eq__,
          '✗ curse:molinari\n  package already installed\n'),
         ('remove curse:molinari',
          str.__eq__,
          '✓ curse:molinari\n  removed\n'),
        ])
    def test_install_with_curse_alias(self, invoke_runner,
                                      test_input, cmp_fn, expected_output):
        assert cmp_fn(invoke_runner(test_input).output, expected_output)

    @pytest.mark.skip(reason='corrupted zip files')
    def test_install_with_tukui_addon_alias(self, invoke_runner):
        assert invoke_runner('install https://www.tukui.org/addons.php?id=3')\
            .output\
            .startswith('✓ tukui:3\n  installed')


    @pytest.mark.skip(reason='corrupted zip files')
    def test_install_with_tukui_ui_alias(self, invoke_runner):
        assert invoke_runner('install https://www.tukui.org/download.php?ui=tukui')\
            .output\
            .startswith('✓ tukui:tukui\n  installed')

    @pytest.mark.parametrize(
        'test_input, cmp_fn, expected_output',
        [('install wowi:10783',
          str.startswith,
          '✓ wowi:10783\n  installed'),
         ('install https://www.wowinterface.com/downloads/info10783-Prat3.0.html',
          str.__eq__,
          '✗ wowi:10783\n  package already installed\n'),
         ('install https://www.wowinterface.com/downloads/landing.php?fileid=10783',
          str.__eq__,
          '✗ wowi:10783\n  package already installed\n'),
         ('remove wowi:10783',
          str.__eq__,
          '✓ wowi:10783\n  removed\n'),
        ])
    def test_install_with_wowi_alias(self, invoke_runner,
                                     test_input, cmp_fn, expected_output):
        assert cmp_fn(invoke_runner(test_input).output, expected_output)




import os
from pathlib import Path
import sys
from unittest.mock import patch

import pytest

from instawow.config import Config
from instawow.exceptions import ConfigError


@pytest.fixture
def folders(tmp_path):
    config = tmp_path / 'config'
    addons = tmp_path / 'addons'
    addons.mkdir()
    return {'config_dir': config, 'addon_dir': addons}


def test_strs_are_coerced_to_paths(folders):
    config = Config(**{k: str(v) for k, v in folders.items()})
    assert config.config_dir == folders['config_dir']
    assert config.addon_dir == folders['addon_dir']


def test_env_vars_have_precedence_over_kwargs(folders, tmp_path):
    config1 = tmp_path / 'config1'
    with patch.dict(os.environ, {'INSTAWOW_CONFIG_DIR': str(config1)}):
        config = Config(**folders)
        assert config.config_dir == config1
        assert config.addon_dir == folders['addon_dir']


def test_config_dir_is_populated(folders):
    config = Config(**folders)
    config.write()
    config_paths = set(config.config_dir.iterdir())
    assert config_paths == {folders['config_dir'] / 'addon_dir.txt',
                            folders['config_dir'] / 'plugins'}


def test_nonexistent_addon_folder_is_rejected(folders, tmp_path):
    addons1 = tmp_path / 'addons1'
    with pytest.raises(ConfigError, match=f"'{addons1}' is not a directory"):
        Config(config_dir=folders['config_dir'], addon_dir=addons1)


def test_reading_addon_folder_from_uninstantiated_profile(folders):
    with pytest.raises(ConfigError, match='configuration not written on disk'):
        Config(config_dir=folders['config_dir'])


def test_reading_addon_folder_from_instantiated_profile(folders):
    Config(**folders).write()
    assert Config(config_dir=folders['config_dir']).addon_dir == folders['addon_dir']


def test_default_config_dir_is_xdg_compliant(folders):
    with patch('sys.platform', 'linux'):
        config_dir = Config(addon_dir=folders['addon_dir']).config_dir
        assert config_dir == Path.home() / '.config/instawow'

        with patch.dict(os.environ, {'XDG_CONFIG_HOME': '/foo'}):
            config_dir = Config(addon_dir=folders['addon_dir']).config_dir
            assert config_dir == Path('/foo/instawow')

    with patch('sys.platform', 'darwin'):
        config_dir = Config(addon_dir=folders['addon_dir']).config_dir
        assert config_dir == Path.home() / 'Library/Application Support/instawow'

        # with patch.dict(os.environ, {'XDG_CONFIG_HOME': 'foo'}):
        #     config_dir = Config(addon_dir=folders['addon_dir']).config_dir
        #     assert config_dir == Path('foo') / 'instawow'




from __future__ import annotations

__all__ = ('main',)

from enum import Enum
from functools import partial, reduce
from textwrap import fill
from typing import (TYPE_CHECKING, Any, Callable, Generator, Iterable, List,
                    NamedTuple, Optional, Sequence, Tuple, Union)

import click

from . import __version__
from .config import Config
from . import exceptions as E
from .manager import CliManager
from .models import Pkg, PkgFolder
from .resolvers import Strategies
from .utils import TocReader, bucketise, is_outdated, setup_logging


class Symbols(str, Enum):

    SUCCESS = click.style('✓', fg='green')
    FAILURE = click.style('✗', fg='red')
    WARNING = click.style('!', fg='blue')

    @classmethod
    def from_result(cls, result: E.ManagerResult) -> Symbols:
        if isinstance(result, E.InternalError):
            return cls.WARNING
        elif isinstance(result, E.ManagerError):
            return cls.FAILURE
        else:
            return cls.SUCCESS


class Report:

    def __init__(self, results: Sequence[Tuple[str, E.ManagerResult]],
                 filter_fn: Callable = (lambda _: True)) -> None:
        self.results = results
        self.filter_fn = filter_fn

    @property
    def code(self) -> int:
        return any(r for _, r in self.results
                   if (isinstance(r, (E.ManagerError, E.InternalError))
                       and self.filter_fn(r)))

    def __str__(self) -> str:
        return '\n'.join(
            (f'{Symbols.from_result(r).value} {click.style(a, bold=True)}\n'
             f'  {r.message}')
            for a, r in self.results
            if self.filter_fn(r)
            )

    def generate_and_exit(self) -> None:
        report = str(self)
        if report:
            click.echo(report)

        ctx = click.get_current_context()
        ctx.exit(self.code)


def tabulate(rows: Iterable, *, show_index: bool = True) -> str:
    from texttable import Texttable, Texttable as c

    table = Texttable(max_width=0).set_deco(c.BORDER | c.HEADER | c.VLINES)
    if show_index:
        rows = iter(rows)
        rows = [('', *next(rows)), *((i, *v) for i, v in enumerate(rows, start=1))]
        table.set_cols_align(('r', *('l' for _ in rows[0]))[:-1])
    else:
        rows = [(), *rows]

    return table.add_rows(rows).draw()


class _Parts(NamedTuple):

    origin: str
    id_or_slug: str


def compose_pkg_uri(val: Any) -> str:
    try:
        origin, slug = val.origin, val.slug
    except AttributeError:
        origin, slug = val
    return ':'.join((origin, slug))


def decompose_pkg_uri(ctx: click.Context,
                      param: click.Parameter,
                      value: Union[str, Tuple[str, ...]],
                      *,
                      raise_for_invalid_uri: bool = True) -> Tuple[str, _Parts]:
    if isinstance(value, tuple):
        return list(bucketise(decompose_pkg_uri(ctx, param, v)
                              for v in value).keys())

    if ':' not in value:
        if raise_for_invalid_uri:
            raise click.BadParameter(value)

        parts = _Parts('*', value)
    else:
        for resolver in ctx.obj.resolvers.values():
            parts = resolver.decompose_url(value)
            if parts:
                parts = _Parts(*parts)
                break
        else:
            parts = _Parts(*value.partition(':')[::2])
    return compose_pkg_uri(parts), parts


def validate_strategy(ctx: click.Context,
                      param: click.Parameter,
                      value: Optional[str]) -> Optional[str]:
    if value and not ctx.params['addons']:
        raise click.UsageError(f'{param.get_error_hint(ctx)} must be used'
                               ' in conjunction with "ADDONS"')
    return value


def get_pkg_from_substr(manager: CliManager, parts: _Parts) -> Optional[Pkg]:
    return manager.get(*parts) or (manager.db.query(Pkg)
                                   .filter(Pkg.slug.contains(parts.id_or_slug))
                                   .order_by(Pkg.name)
                                   .first())


class _OrigCmdOrderGroup(click.Group):

    def list_commands(self, ctx: click.Context) -> List[str]:
        return list(self.commands)    # The default is ``sorted(self.commands)``


def create_config() -> Config:
    """Create the configuration.  This prompts the user for their add-on folder
    on first run or if the folder can't be found.
    """
    try:
        return Config().write()
    except E.ConfigError as error:
        # readline is picked up by ``input`` to undumbify line editing
        import readline

        # Don't bother if Python was built without GNU readline - we'd have to
        # reimplement path completion
        if 'GNU readline' in getattr(readline, '__doc__', ''):
            readline.parse_and_bind('tab: complete')
            readline.set_completer_delims('')   # Do not split up the string

        message = (
            f'{Symbols.WARNING.value} {{.message}}\n'
            f'  {click.style(">", fg="yellow")} enter the path to your add-on folder: ')

        def prompt(error: E.ConfigError) -> Config:
            try:
                addon_dir = input(message.format(error))
            except E.ConfigError as error:
                return prompt(error)
            else:
                return Config(addon_dir=addon_dir).write()

        return prompt(error)


@click.group(cls=_OrigCmdOrderGroup,
             context_settings={'help_option_names': ['-h', '--help']})
@click.version_option(__version__, prog_name='instawow')
@click.pass_context
def main(ctx):
    "Add-on manager for World of Warcraft."
    try:
        import uvloop
    except ImportError:
        pass
    else:
        import asyncio
        asyncio.set_event_loop_policy(uvloop.EventLoopPolicy())

    if not ctx.obj and ctx.invoked_subcommand != 'serve':
        config = create_config()
        setup_logging(config)

        if ctx.invoked_subcommand == 'extras':
            ctx.obj = CliManager(config, show_progress=False)
        else:
            ctx.obj = manager = CliManager(config)
            if is_outdated(manager):
                click.echo(f'{Symbols.WARNING.value} instawow is out of date')


@main.command()
@click.argument('addons', nargs=-1, required=True, callback=decompose_pkg_uri)
@click.option('--strategy', '-s',
              type=click.Choice({s.value for s in Strategies}),
              default='default',
              help="Whether to install the latest published version "
                   "('default') or the very latest upload ('latest').")
@click.option('--replace', '-o',
              is_flag=True, default=False,
              help='Replace existing add-ons.')
@click.pass_obj
def install(manager, addons, strategy, replace) -> None:
    "Install add-ons."
    results = zip((d for d, _ in addons),
                  manager.install_many((*p, strategy, replace) for _, p in addons))
    Report(list(results)).generate_and_exit()


@main.command()
@click.argument('addons', nargs=-1, callback=decompose_pkg_uri, is_eager=True)
@click.option('--strategy', '-s',
              callback=validate_strategy,
              type=click.Choice({s.value for s in Strategies}),
              default=None,
              help="Whether to update to the latest published version "
                   "('default') or the very latest upload ('latest').")
@click.pass_obj
def update(manager, addons, strategy) -> None:
    "Update installed add-ons."
    orig_addons = addons
    if not addons:
        addons = [(compose_pkg_uri(p), (p.origin, p.slug))
                  for p in manager.db.query(Pkg).all()]

    if strategy:
        with manager.db.begin_nested():
            for pkg in (manager.get(*p) for _, p in addons):
                pkg.options.strategy = strategy

    results = zip((d for d, _ in addons),
                  manager.update_many(p for _, p in addons))
    # Hide if ``update`` was invoked without arguments and the package is
    # up-to-date or temporarily unavailable
    filter_fn = lambda r: (orig_addons
                           or not isinstance(r, (E.PkgUpToDate, E.PkgTemporarilyUnavailable)))
    Report(list(results), filter_fn).generate_and_exit()


@main.command()
@click.argument('addons', nargs=-1, required=True, callback=decompose_pkg_uri)
@click.pass_obj
def remove(manager, addons) -> None:
    "Uninstall add-ons."
    def remove_():
        for addon, parts in addons:
            try:
                yield (addon, manager.run(manager.remove(*parts)))
            except (E.ManagerError, E.InternalError) as error:
                yield (addon, error)

    Report(list(remove_())).generate_and_exit()


@main.command('list')
@click.option('--column', '-c', 'columns',
              multiple=True,
              help='A field to show in a column.  Nested fields are '
                   'dot-delimited.  Repeatable.')
@click.option('--columns', '-C', 'print_columns',
              is_flag=True, default=False,
              help='Print a list of all possible column values.')
@click.option('--toc-entry', '-t', 'toc_entries',
              multiple=True,
              help='An entry to extract from the TOC.  Repeatable.')
@click.option('--sort-by', '-s', 'sort_key',
              default='name',
              help='A key to sort the table by.  '
                   'You can chain multiple keys by separating them with a comma '
                   'just as you would in SQL, '
                   'e.g. `--sort-by="origin, date_published DESC"`.')
@click.pass_obj
def list_installed(manager, columns, print_columns, toc_entries, sort_key) -> None:
    "List installed add-ons."
    from sqlalchemy import inspect, text

    def format_columns(pkg):
        for column in columns:
            try:
                value = reduce(getattr, [pkg] + column.split('.'))
            except AttributeError:
                raise click.BadParameter(column, param_hint=['--column', '-c'])
            if column == 'folders':
                value = '\n'.join(f.name for f in value)
            elif column == 'options':
                value = f'strategy = {value.strategy}'
            elif column == 'description':
                value = fill(value, width=50, max_lines=3)
            yield value

    def format_toc_entries(pkg):
        if toc_entries:
            toc_readers = [TocReader(manager.config.addon_dir
                                     / f.name
                                     / f'{f.name}.toc') for f in pkg.folders]
            for toc_entry in toc_entries:
                value = [fill(r[toc_entry].value, width=50) for r in toc_readers]
                value = sorted(set(value), key=value.index)
                yield '\n'.join(value)

    if print_columns:
        rows = [('field',),
                *((c,) for c in (*inspect(Pkg).columns.keys(),
                                 *inspect(Pkg).relationships.keys()))]
        click.echo(tabulate(rows, show_index=False))
    else:
        pkgs = manager.db.query(Pkg).order_by(text(sort_key)).all()
        if pkgs:
            rows = [('add-on', *columns, *(f'[{e}]' for e in toc_entries)),
                     *((compose_pkg_uri(p), *format_columns(p), *format_toc_entries(p))
                    for p in pkgs)]
            click.echo(tabulate(rows))


@main.command()
@click.pass_obj
def list_uncontrolled(manager) -> None:
    "List add-ons not managed by instawow."
    folders = ({f for f in manager.config.addon_dir.iterdir() if f.is_dir()}
               - {manager.config.addon_dir / f.name
                  for f in manager.db.query(PkgFolder).all()})
    folders_and_tocs = ((n, n / f'{n.name}.toc') for n in folders)
    folders_and_readers = sorted((n, TocReader(t))
                                 for n, t in folders_and_tocs if t.exists())
    if folders:
        rows = [('folder', 'Curse ID', 'WoWI ID', 'version'),
                *((n.name,
                   t['X-Curse-Project-ID'].value,
                   t['X-WoWI-ID'].value,
                   t['X-Curse-Packaged-Version', 'X-Packaged-Version', 'Version'].value)
                  for n, t in folders_and_readers)]
        click.echo(tabulate(rows))


@main.command()
@click.argument('addon', callback=partial(decompose_pkg_uri,
                                          raise_for_invalid_uri=False))
@click.option('--toc-entry', '-t', 'toc_entries',
              multiple=True,
              help='An entry to extract from the TOC.  Repeatable.')
@click.pass_obj
def info(manager, addon, toc_entries) -> None:
    "Show the details of an installed add-on."
    pkg = get_pkg_from_substr(manager, addon[1])
    if pkg:
        rows = {'name': pkg.name,
                'source': pkg.origin,
                'id': pkg.id,
                'slug': pkg.slug,
                'description': fill(pkg.description, max_lines=5),
                'homepage': pkg.url,
                'version': pkg.version,
                'release date': pkg.date_published,
                'folders': '\n'.join([str(manager.config.addon_dir)]
                                     + [' ├─ ' + f.name for f in pkg.folders[:-1]]
                                     + [' └─ ' + pkg.folders[-1].name]),
                'strategy': pkg.options.strategy}

        if toc_entries:
            for folder in pkg.folders:
                toc_reader = TocReader(manager.config.addon_dir
                                       / folder.name
                                       / f'{folder.name}.toc')
                rows.update({f'[{folder.name} {k}]': fill(toc_reader[k].value)
                             for k in toc_entries})
        click.echo(tabulate(rows.items(), show_index=False))
    else:
        Report([(addon[0], E.PkgNotInstalled())]).generate_and_exit()


@main.command()
@click.argument('addon', callback=partial(decompose_pkg_uri,
                                          raise_for_invalid_uri=False))
@click.pass_obj
def visit(manager, addon) -> None:
    "Open the add-on's homepage in your browser."
    pkg = get_pkg_from_substr(manager, addon[1])
    if pkg:
        import webbrowser
        webbrowser.open(pkg.url)
    else:
        Report([(addon[0], E.PkgNotInstalled())]).generate_and_exit()


@main.command()
@click.argument('addon', callback=partial(decompose_pkg_uri,
                                          raise_for_invalid_uri=False))
@click.pass_obj
def reveal(manager, addon) -> None:
    "Open the add-on folder in your file manager."
    pkg = get_pkg_from_substr(manager, addon[1])
    if pkg:
        import webbrowser
        webbrowser.open((manager.config.addon_dir / pkg.folders[0].name).as_uri())
    else:
        Report([(addon[0], E.PkgNotInstalled())]).generate_and_exit()


@main.command()
@click.option('--port', type=int, help='The server port.')
def serve(port) -> None:
    "Run the WebSocket server."
    from .manager import WsManager
    WsManager().serve(port=port)


@main.group('extras',
            cls=_OrigCmdOrderGroup)
def _extras_group() -> None:
    "Additional functionality."


@_extras_group.group('weakauras',
                     cls=_OrigCmdOrderGroup)
def _weakauras_group() -> None:
    "Manage your WeakAuras."


@_extras_group.group('bitbar',
                     cls=_OrigCmdOrderGroup)
def _bitbar_group() -> None:
    """Mini update GUI implemented as a BitBar plug-in.

    BitBar <https://getbitbar.com/> is a menu-bar app host for macOS.
    """


@_bitbar_group.command('_generate', hidden=True)
@click.argument('caller')
@click.argument('version')
@click.pass_obj
def generate_bitbar_output(manager, caller, version):
    from base64 import b64encode
    from pathlib import Path
    from subprocess import run, PIPE

    installed = manager.db.query(Pkg).order_by(Pkg.name).all()
    latest = manager.resolve_many((p.origin, p.id, p.options.strategy) for p in installed)
    outdated = [(l, r) for l, r in zip(installed, latest)
                if l.file_id != getattr(r, 'file_id', l.file_id)]

    icon = Path(__file__).parent / 'assets' / f'''\
NSStatusItem-icon__\
{"has-updates" if outdated else "clear"}\
{"@2x" if b"Retina" in run(("system_profiler", "SPDisplaysDataType"),
                           stdout=PIPE).stdout else ""}\
.png'''
    icon = b64encode(icon.read_bytes()).decode()

    click.echo(f'''| templateImage="{icon}"
---
instawow ({__version__}:{version})''')
    if outdated:
        if len(outdated) > 1:
            click.echo(f'''\
Update all | bash={caller} param1=update terminal=false refresh=true''')
        for p, r in outdated:
            click.echo(f'''\
Update “{r.name}” ({p.version} ➞ {r.version}) | \
  refresh=true \
  bash={caller} param1=update param2={_compose_pkg_uri(r)} terminal=false
View “{r.name}” on {manager.resolvers[r.origin].name} | \
  alternate=true \
  bash={caller} param1=hearth param2={_compose_pkg_uri(r)} terminal=false''')


@_bitbar_group.command('install')
def install_bitbar_plugin():
    "Install the instawow BitBar plug-in."
    from pathlib import Path
    import sys
    import tempfile
    import webbrowser

    with tempfile.TemporaryDirectory() as name:
        path = Path(name, 'instawow.1h.py')
        path.write_text(f'''\
#!/usr/bin/env LC_ALL=en_US.UTF-8 {sys.executable}

__version__ = {__version__!r}

import sys

from instawow.cli import main

main(['-n', *(sys.argv[1:] or ['extras', 'bitbar', '_generate', sys.argv[0], __version__])])
''')
        webbrowser.open(f'bitbar://openPlugin?src={path.as_uri()}')
        click.pause('Press any key to exit after installing the plug-in')


@_weakauras_group.command('build-companion')
@click.option('--account', '-a',
              required=True,
              help='Your account name.  This is used to locate '
                   'the WeakAuras data file.')
@click.pass_obj
def build_weakauras_companion(manager, account) -> None:
    "Build the WeakAuras Companion add-on."
    from .wa_updater import WaCompanionBuilder
    WaCompanionBuilder(manager).build(account)


@_weakauras_group.command('list')
@click.option('--account', '-a',
              required=True,
              help='Your account name.  This is used to locate '
                   'the WeakAuras data file.')
@click.pass_obj
def list_installed_wago_auras(manager, account) -> None:
    "List WeakAuras installed from Wago."
    from .wa_updater import WaCompanionBuilder

    aura_groups = WaCompanionBuilder(manager).extract_installed_auras(account)
    installed_auras = [(a.url, str(a.ignore_wago_update).lower())
                       for a, *_ in aura_groups.values()]
    click.echo(tabulate([('url', 'ignore updates'), *installed_auras]))




import pytest

from instawow.config import Config
from instawow.manager import Manager
from instawow.wa_updater import (AuraEntry, ApiMetadata, WaCompanionBuilder,
                                 URL, bucketise)


@pytest.fixture
def builder(tmp_path):
    addons = tmp_path / 'World of Warcraft/Interface/AddOns'
    addons.mkdir(parents=True)
    config = Config(config_dir=tmp_path / 'config', addon_dir=addons)
    config.write()
    yield WaCompanionBuilder(Manager(config))


@pytest.fixture
def event_loop(builder):
    yield builder.loop


def test_bucketise_bucketises_by_putting_things_in_a_bucketing_bucket():
    assert bucketise(iter([1, 1, 0, 1]), bool) == {True: [1, 1, 1], False: [0]}


@pytest.mark.asyncio
@pytest.mark.parametrize('ids', [['RaidCDs', 'bfaraid2'],
                                 ['bfaraid2', 'RaidCDs'],])
async def test_id_order_is_retained_in_aura_metadata(builder, ids):
    builder.client.set(await builder.client_factory())
    results = await builder.get_wago_aura_metadata(ids)
    assert ids == [r.slug for r in results]


@pytest.mark.asyncio
async def test_id_length_is_retained_in_aura_metadata(builder):
    builder.client.set(await builder.client_factory())
    ids = ['bfaraid2', 'foobar', 'RaidCDs']    # Invalid ID flanked by two valid IDs
    results = await builder.get_wago_aura_metadata(ids)
    assert [ApiMetadata, type(None), ApiMetadata] == [type(r) for r in results]


def test_can_parse_empty_displays_table(builder):
    assert builder.extract_auras_from_lua('''\
WeakAurasSaved = {
    ["displays"] = {
    },
}
''') == {}


def test_urlless_display_is_discarded(builder):
    assert builder.extract_auras_from_lua('''\
WeakAurasSaved = {
    ["displays"] = {
        ["Foo"] = {
            ["bar"] = "baz",
        },
    },
}
''') == {}


def test_can_parse_minimal_wago_display(builder):
    aura = {'id': 'foo',
            'uid': 'foo',
            'parent': None,
            'url': URL('https://wago.io/foo/1'),
            'version': 1,
            'semver': None,
            'ignore_wago_update': False}
    assert builder.extract_auras_from_lua('''\
WeakAurasSaved = {
    ["displays"] = {
        ["Foo"] = {
            ["bar"] = "baz",
            ["url"] = "https://wago.io/foo/1",
            ["version"] = 1,
            ["id"] = "foo",
            ["uid"] = "foo",
        },
    },
}
''') == {'foo': [AuraEntry.construct(aura, set(aura))]}


def test_url_host_not_wago_display_is_discarded(builder):
    assert builder.extract_auras_from_lua('''\
WeakAurasSaved = {
    ["displays"] = {
        ["Foo"] = {
            ["bar"] = "baz",
            ["url"] = "https://wafo.io/foo/1",
            ["version"] = 1,
            ["id"] = "foo",
            ["uid"] = "foo",
        },
    },
}
''') == {}


def test_can_build_addon_without_updates(builder, request):
    builder.builder_dir.mkdir()
    builder.make_addon([])
    request.config.cache.set('orig_checksum', builder.checksum())


def test_build_is_reproducible(builder, request):
    builder.builder_dir.mkdir()
    builder.make_addon([])
    assert request.config.cache.get('orig_checksum', None) == builder.checksum()




from __future__ import annotations

__all__ = ('ErrorCodes',
           'ApiError',
           'ResolveRequest',
           'InstallRequest',
           'UpdateRequest',
           'RemoveRequest',
           'GetRequest',
           'SuccessResponse',
           'ErrorResponse',
           'jsonify',
           'parse_request')

import abc
from datetime import datetime
from enum import Enum, IntEnum
from functools import partial
from json import JSONDecodeError, JSONEncoder, loads
from pathlib import Path
from typing import (Any, Awaitable, Callable, ClassVar, Dict, Generator,
                    List, Optional, Tuple, TypeVar, Union)

from pydantic import BaseModel, Extra, StrictStr, ValidationError, validator
from pydantic.errors import IntegerError

from .config import Config
from . import exceptions as E
from .manager import WsManager
from .models import *
from .resolvers import Strategies
from .utils import setup_logging


JSONRPC_VERSION = '2.0'


class ErrorCodes(IntEnum):

    PARSE_ERROR                     = -32700
    INVALID_REQUEST                 = -32600
    METHOD_NOT_FOUND                = -32601
    INVALID_PARAMS                  = -32602
    INTERNAL_ERROR                  = -32603

    MANAGER_ERROR                   = +10000
    CONFIG_ERROR                    = +10010
    PKG_ALREADY_INSTALLED           = +10021
    PKG_CONFLICTS_WITH_INSTALLED    = +10022
    PKG_CONFLICTS_WITH_UNCONTROLLED = +10023
    PKG_NONEXISTENT                 = +10024
    PKG_TEMPORARILY_UNAVAILABLE     = +10025
    PKG_NOT_INSTALLED               = +10026
    PKG_ORIGIN_INVALID              = +10027
    PKG_UP_TO_DATE                  = +10028
    PKG_STRATEGY_INVALID            = +10029

    ManagerError                    = +10000
    ConfigError                     = +10010
    PkgAlreadyInstalled             = +10021
    PkgConflictsWithInstalled       = +10022
    PkgConflictsWithUncontrolled    = +10023
    PkgNonexistent                  = +10024
    PkgTemporarilyUnavailable       = +10025
    PkgNotInstalled                 = +10026
    PkgOriginInvalid                = +10027
    PkgUpToDate                     = +10028
    PkgStrategyInvalid              = +10029


class _ApiErrorMeta(type):

    def __getattr__(cls, name: str) -> Callable:
        return partial(ApiError, ErrorCodes[name])


class ApiError(Exception, metaclass=_ApiErrorMeta):

    def __init__(self,
                 error_code: ErrorCodes,
                 message: str,
                 data: Optional[str]=None,
                 request_id: Union[None, int, str]=None) -> None:
        self.error_code = error_code
        self.message = message
        self.data = data
        self.request_id = request_id


class _IncompatibleMethodError(ValueError):
    pass


def validate_const(const):
    "A validator mirroring JSON Schema's ``const``."
    def is_equal(cls, value):
        if value != const:
            raise _IncompatibleMethodError
        return value

    is_equal.__qualname__ = f'is_equal_{id(is_equal)}'    # Needed to fool pydantic's validator registry
    return is_equal


def decompose_pkg_uri(manager: WsManager, uri: str) -> Tuple[str, str]:
    "Turn a URI into an origin–slug pair."
    resolvers = manager.resolvers.values()
    try:
        return next(filter(None, (r.decompose_url(uri) for r in resolvers)))
    except StopIteration:
        return uri.partition(':')[::2]


class Message(BaseModel,
              abc.ABC):

    jsonrpc: str

    __is_jsonrpc_const = validator('jsonrpc')(validate_const(JSONRPC_VERSION))


class StrictInt(int):

    @classmethod
    def __get_validators__(cls) -> Generator:
        yield cls.validate

    @classmethod
    def validate(cls, v: Any) -> int:
        if type(v) is not int:      # isintance(bool(), int) == True
            raise IntegerError
        return v


class RequestParams(BaseModel):
    pass


class Request(Message,
              abc.ABC):

    _name: ClassVar[str]

    id: Union[StrictInt, StrictStr]
    method: str
    params: RequestParams

    def prepare_response(self, manager: WsManager) -> Awaitable:
        raise NotImplementedError

    def consume_result(self, result: Any) -> SuccessResponse:
        raise NotImplementedError


class BaseRequestParams(RequestParams):

    class Config:
        extra = Extra.allow


class BaseRequest(Request):

    _name = '!base_request'

    params: BaseRequestParams = BaseRequestParams()


class SetupRequestParams(RequestParams):

    addon_dir: Path


class SetupRequest(Request):

    _name = 'setup'

    params: SetupRequestParams

    __is_method_const = validator('method')(validate_const('setup'))

    def prepare_response(self, manager: WsManager) -> Awaitable:
        async def setup() -> Config:
            config = Config(addon_dir=self.params.addon_dir).write()
            setup_logging(config)
            manager.finalise(config)
            return config

        return setup()

    def consume_result(self, result: Config) -> SuccessResponse:
        return SuccessResponse(result={'config': result.__dict__}, id=self.id)


class ResolveRequestParams(RequestParams):

    uri: str
    resolution_strategy: Strategies

    class Config:
        use_enum_values = True


class ResolveRequest(Request):

    _name = 'resolve'

    params: ResolveRequestParams

    __is_method_const = validator('method')(validate_const('resolve'))

    def prepare_response(self, manager: WsManager) -> Awaitable:
        return manager.resolve(*decompose_pkg_uri(manager, self.params.uri),
                               self.params.resolution_strategy)

    def consume_result(self, result: Any) -> SuccessResponse:
        return SuccessResponse(result=result, id=self.id)


class InstallRequestParams(RequestParams):

    uri: str
    resolution_strategy: Strategies
    replace: bool

    class Config:
        use_enum_values = True


class InstallRequest(Request):

    _name = 'install'

    params: InstallRequestParams

    __is_method_const = validator('method')(validate_const('install'))

    def prepare_response(self, manager: WsManager) -> Awaitable:
        return manager.to_install(*decompose_pkg_uri(manager, self.params.uri),
                                  self.params.resolution_strategy,
                                  self.params.replace)

    def consume_result(self, result: Any) -> SuccessResponse:
        return SuccessResponse(result=result.new_pkg, id=self.id)


class UpdateRequestParams(RequestParams):

    uri: str


class UpdateRequest(Request):

    _name = 'update'

    params: UpdateRequestParams

    __is_method_const = validator('method')(validate_const('update'))

    def prepare_response(self, manager: WsManager) -> Awaitable:
        return manager.to_update(*decompose_pkg_uri(manager, self.params.uri))

    def consume_result(self, result: Any) -> SuccessResponse:
        return SuccessResponse(result=result.new_pkg, id=self.id)


class RemoveRequestParams(RequestParams):

    uri: str


class RemoveRequest(Request):

    _name = 'remove'

    params: RemoveRequestParams

    __is_method_const = validator('method')(validate_const('remove'))

    def prepare_response(self, manager: WsManager) -> Awaitable:
        return manager.remove(*decompose_pkg_uri(manager, self.params.uri))

    def consume_result(self, result: Any) -> SuccessResponse:
        return SuccessResponse(result=None, id=self.id)


class GetRequestParams(RequestParams):

    uris: List[str]


class GetRequest(Request):

    _name = 'get'

    params: GetRequestParams

    __is_method_const = validator('method')(validate_const('get'))

    def prepare_response(self, manager: WsManager) -> Awaitable:
        async def get() -> list:
            if self.params.uris:
                return [manager.get(*decompose_pkg_uri(manager, u))
                        for u in self.params.uris]
            return manager.db.query(Pkg).all()

        return get()

    def consume_result(self, result: Any) -> SuccessResponse:
        return SuccessResponse(result=result, id=self.id)


class Response(Message,
               abc.ABC):

    jsonrpc = JSONRPC_VERSION
    id: Union[None, StrictInt, StrictStr]


class SuccessResponse(Response):

    result: Any


class Error(BaseModel):

    code: ErrorCodes
    message: str
    data: Optional[str]

    class Config:
        use_enum_values = True


class ErrorResponse(Response):

    error: Error

    @classmethod
    def from_api_error(cls, api_error: ApiError) -> ErrorResponse:
        return cls(id=api_error.request_id, error=Error(code=api_error.error_code,
                                                        message=api_error.message,
                                                        data=api_error.data))


def _convert_sqla_obj(obj: Union[Pkg, PkgFolder, PkgOptions]) -> dict:
    models = {Pkg: (PkgCoercer, [*PkgCoercer.__fields__, 'folders', 'options']),
              PkgFolder: (PkgFolderCoercer, ['path']),
              PkgOptions: (PkgOptionsCoercer, ['strategy'])}

    coercer, fields = models[obj.__class__]
    return coercer.parse_obj({f: getattr(obj, f) for f in fields}).dict()


_CONVERTERS = [(Pkg, _convert_sqla_obj),
               (PkgFolder, _convert_sqla_obj),
               (PkgOptions, _convert_sqla_obj),
               (Message, Message.dict),
               (Path, str),
               (datetime, datetime.isoformat),]


class Encoder(JSONEncoder):

    def default(self, value: Any) -> Any:
        try:
            return next(c(value) for t, c in _CONVERTERS if isinstance(value, t))
        except StopIteration:
            return super().default(value)


jsonify = Encoder().encode


TR = TypeVar('TR', bound=Request)

_REQUESTS = {r._name: r for r in {SetupRequest,
                                  ResolveRequest,
                                  InstallRequest,
                                  UpdateRequest,
                                  RemoveRequest,
                                  GetRequest}}


def parse_request(message: str) -> TR:
    "Parse a JSON string into a sub-``Request`` object."
    try:
        base_request = BaseRequest.parse_obj(loads(message))
    except JSONDecodeError as error:
        raise ApiError.PARSE_ERROR('request is not valid JSON', str(error))
    except ValidationError as error:
        raise ApiError.INVALID_REQUEST('request is malformed',
                                       error.json(indent=None))
    try:
        return _REQUESTS[base_request.method].parse_obj(base_request.dict())
    except KeyError:
        raise ApiError.METHOD_NOT_FOUND('request method not found',
                                        None, base_request.id)
    except ValidationError as error:
        raise ApiError.INVALID_PARAMS('request params are invalid',
                                      error.json(indent=None), base_request.id)




import asyncio
from multiprocessing import Process
import os
from typing import List
from unittest.mock import patch

import aiohttp
import pytest

from instawow import api
from instawow.config import Config
from instawow.manager import WsManager
from instawow.models import PkgCoercer, PkgFolderCoercer, PkgOptionsCoercer


PORT = 55439


class Pkg(PkgCoercer):

    folders: List[PkgFolderCoercer]
    options: PkgOptionsCoercer


class FailRequest(api.Request):

    _name = 'fail'

    def prepare_response(self, manager):
        async def raise_():
            raise ValueError

        return raise_()


@pytest.fixture(scope='module', autouse=True)
def config(tmp_path_factory):
    config = Config(config_dir=tmp_path_factory.mktemp(f'{__name__}_config'),
                    addon_dir= tmp_path_factory.mktemp(f'{__name__}_addons'))
    config.write()
    yield config


@pytest.fixture(scope='module', autouse=True)
def environ(config):
    with patch.dict(os.environ, {'INSTAWOW_CONFIG_DIR': str(config.config_dir)}):
        yield



@pytest.fixture(scope='module', autouse=True)
def ws_server(config):
    with patch.dict(api._REQUESTS, {FailRequest._name: FailRequest}):
        process = Process(daemon=True, target=lambda: WsManager().serve(port=PORT))
        process.start()
        process.join(3)         # Effectively ``time.sleep(3)``
        yield


@pytest.fixture
@pytest.mark.asyncio
async def ws_client(ws_server):
    async with aiohttp.ClientSession() as client, \
            client.ws_connect(f'ws://127.0.0.1:{PORT}') as ws:
        yield ws


@pytest.mark.asyncio
async def test_parse_error_error_response(ws_client):
    request = 'foo'
    response = {'jsonrpc': '2.0',
                'id': None,
                'error': {'code': -32700,
                          'message': 'request is not valid JSON',
                          'data': 'Expecting value: line 1 column 1 (char 0)'}}
    await ws_client.send_str('foo')
    assert (await ws_client.receive_json()) == response


@pytest.mark.asyncio
async def test_invalid_request_error_response(ws_client):
    request = []
    response = {'jsonrpc': '2.0',
                'id': None,
                'error': {'code': -32600,
                          'message': 'request is malformed',
                          'data': '[{"loc": ["__obj__"], "msg": "BaseRequest expected dict not '
                                  'list", "type": "type_error"}]'}}
    await ws_client.send_json(request)
    assert (await ws_client.receive_json()) == response


@pytest.mark.asyncio
async def test_method_not_found_error_response(ws_client):
    request = {'jsonrpc': '2.0',
               'id': 'test_method_not_found_error_response',
               'method': 'foo'}
    response = {'jsonrpc': '2.0',
                'id': 'test_method_not_found_error_response',
                'error': {'code': -32601,
                          'message': 'request method not found',
                          'data': None}}
    await ws_client.send_json(request)
    assert (await ws_client.receive_json()) == response


@pytest.mark.asyncio
async def test_invalid_params_error_response(ws_client):
    request = {'jsonrpc': '2.0',
               'id': 'test_invalid_params_error_response',
               'method': 'get',
               'params': {}}
    response = {'jsonrpc': '2.0',
                'id': 'test_invalid_params_error_response',
                'error': {'code': -32602,
                          'message': 'request params are invalid',
                          'data': '[{"loc": ["params", "uris"], "msg": "field required", '
                                  '"type": "value_error.missing"}]'}}
    await ws_client.send_json(request)
    assert (await ws_client.receive_json()) == response


@pytest.mark.asyncio
async def test_internal_error_error_response(ws_client):
    request = {'jsonrpc': '2.0',
               'id': 'test_internal_error_error_response',
               'method': FailRequest._name,
               'params': {}}
    response = {'jsonrpc': '2.0',
                'id': 'test_internal_error_error_response',
                'error': {'code': -32603,
                          'message': 'encountered an internal error',
                          'data': None}}
    await ws_client.send_json(request)
    assert (await ws_client.receive_json()) == response


@pytest.mark.asyncio
async def test_setup_request_response(ws_client, config):
    request = {'jsonrpc': '2.0',
               'id': 'test_setup_request_response',
               'method': 'setup',
               'params': {'addon_dir': str(config.addon_dir)}}
    response = {'jsonrpc': '2.0',
                'id': 'test_setup_request_response',
                'result': {'config': {k: str(v) for k, v in config.__dict__.items()}}}
    await ws_client.send_json(request)
    assert (await ws_client.receive_json()) == response


@pytest.mark.asyncio
async def test_any_manager_error_error_response(ws_client):
    request = {'jsonrpc': '2.0',
               'id': 'test_any_manager_error_error_response',
               'method': 'resolve',
               'params': {'uri': 'foo:bar', 'resolution_strategy': 'default'}}
    response = {'jsonrpc': '2.0',
                'id': 'test_any_manager_error_error_response',
                'error': {'code': 10027,
                          'message': 'package origin is invalid',
                          'data': None}}
    await ws_client.send_json(request)
    assert (await ws_client.receive_json()) == response


@pytest.mark.asyncio
async def test_get_method_not_installed_response(ws_client):
    request = {'jsonrpc': '2.0',
               'id': 'test_get_method_not_installed_response',
               'method': 'get',
               'params': {'uris': ['curse:molinari']}}

    await ws_client.send_json(request)
    response = await ws_client.receive_json()
    assert response['id'] == request['id']
    assert api.SuccessResponse.parse_obj(response).result == [None]


@pytest.mark.asyncio
async def test_resolve_method_response(ws_client):
    request = {'jsonrpc': '2.0',
               'id': 'test_resolve_method_response',
               'method': 'resolve',
               'params': {'uri': 'curse:molinari', 'resolution_strategy': 'default'}}

    await ws_client.send_json(request)
    response = await ws_client.receive_json()
    assert response['id'] == request['id']
    api.SuccessResponse.parse_obj(response)
    pkg = Pkg.parse_obj(response['result'])
    assert pkg.folders == []


@pytest.mark.asyncio
async def test_resolve_method_with_url_in_uri_param_response(ws_client):
    request = {'jsonrpc': '2.0',
               'id': 'test_resolve_method_with_url_in_uri_param_response',
               'method': 'resolve',
               'params': {'uri': 'https://www.curseforge.com/wow/addons/molinari',
                          'resolution_strategy': 'default'}}

    await ws_client.send_json(request)
    response = await ws_client.receive_json()
    assert response['id'] == request['id']
    assert response['result']['origin'] == 'curse'
    assert response['result']['slug'] == 'molinari'


@pytest.mark.asyncio
async def test_install_method_response(ws_client):
    request = {'jsonrpc': '2.0',
               'id': 'test_install_method_response',
               'method': 'install',
               'params': {'uri': 'curse:molinari', 'resolution_strategy': 'default',
                          'overwrite': False}}

    await ws_client.send_json(request)
    response = api.SuccessResponse.parse_obj(await ws_client.receive_json())
    assert response.id == request['id']
    Pkg.parse_obj(response.result)


@pytest.mark.asyncio
async def test_get_method_response(ws_client):
    request = {'jsonrpc': '2.0',
               'id': 'test_get_method_response',
               'method': 'get',
               'params': {'uris': ['curse:molinari']}}

    await ws_client.send_json(request)
    response = await ws_client.receive_json()
    assert response['id'] == request['id']
    assert len(api.SuccessResponse.parse_obj(response).result) == 1
    pkg = Pkg.parse_obj(response['result'][0])
    assert len(pkg.folders) > 0


@pytest.mark.asyncio
async def test_remove_method_response(ws_client):
    request = {'jsonrpc': '2.0',
               'id': 'test_remove_method_response',
               'method': 'remove',
               'params': {'uri': 'curse:molinari'}}

    await ws_client.send_json(request)
    response = api.SuccessResponse.parse_obj(await ws_client.receive_json())
    assert response.id == request['id']
    assert response.result is None




from __future__ import annotations

__all__ = ('ManagerResult',
           'PkgInstalled',
           'PkgUpdated',
           'PkgRemoved',
           'ManagerError',
           'ConfigError',
           'PkgAlreadyInstalled',
           'PkgConflictsWithInstalled',
           'PkgConflictsWithUncontrolled',
           'PkgNonexistent',
           'PkgTemporarilyUnavailable',
           'PkgNotInstalled',
           'PkgOriginInvalid',
           'PkgUpToDate',
           'PkgStrategyInvalid',
           'InternalError')

from typing import TYPE_CHECKING, ClassVar, Set

if TYPE_CHECKING:
    from .models import Pkg


class ManagerResult:

    fmt_message: ClassVar[str]

    def __call__(self) -> ManagerResult:
        return self

    @property
    def message(self) -> str:
        return self.fmt_message.format(self=self)


class PkgInstalled(ManagerResult):

    fmt_message = 'installed {self.new_pkg.version}'

    def __init__(self, new_pkg: Pkg) -> None:
        super().__init__()
        self.new_pkg = new_pkg


class PkgUpdated(ManagerResult):

    fmt_message = 'updated {self.old_pkg.version} to {self.new_pkg.version}'

    def __init__(self, old_pkg: Pkg, new_pkg: Pkg) -> None:
        super().__init__()
        self.old_pkg = old_pkg
        self.new_pkg = new_pkg


class PkgRemoved(ManagerResult):

    fmt_message = 'removed'

    def __init__(self, old_pkg: Pkg) -> None:
        super().__init__()
        self.old_pkg = old_pkg


class ManagerError(ManagerResult,
                   Exception):
    pass


class ConfigError(ManagerError):

    fmt_message = '{self._message}'

    def __init__(self, message: str) -> None:
        super().__init__(message)
        self._message = message


class PkgAlreadyInstalled(ManagerError):

    fmt_message = 'package already installed'


class PkgConflictsWithInstalled(ManagerError):

    fmt_message = "package folders conflict with installed package's "\
                  '{self.conflicting_pkg.origin}:{self.conflicting_pkg.slug}'

    def __init__(self, conflicting_pkg: Pkg) -> None:
        super().__init__()
        self.conflicting_pkg = conflicting_pkg


class PkgConflictsWithUncontrolled(ManagerError):

    fmt_message = "package folders conflict with an add-on's"\
                  ' not controlled by instawow'

    def __init__(self, folders: Set[str]) -> None:
        super().__init__()
        self.folders = folders


class PkgNonexistent(ManagerError):

    fmt_message = 'package does not exist'


class PkgTemporarilyUnavailable(ManagerError):

    fmt_message = 'package is temporarily unavailable'


class PkgNotInstalled(ManagerError):

    fmt_message = 'package is not installed'


class PkgOriginInvalid(ManagerError):

    fmt_message = 'package origin is invalid'


class PkgUpToDate(ManagerError):

    fmt_message = 'package is up to date'


class PkgStrategyInvalid(ManagerError):

    fmt_message = '{self.strategy!r} is not a valid strategy'

    def __init__(self, strategy: str) -> None:
        super().__init__()
        self.strategy = strategy


class InternalError(ManagerResult,
                    Exception):

    fmt_message = 'encountered {self.error.__class__.__name__}'

    def __init__(self, error: BaseException) -> None:
        super().__init__()
        self.error = error




from __future__ import annotations

__all__ = ('Manager', 'CliManager', 'WsManager')

import asyncio
from contextlib import asynccontextmanager
import contextvars as cv
from functools import partial
from pathlib import Path, PurePath
from shutil import move as _move
from tempfile import NamedTemporaryFile, mkdtemp
from typing import *
from zipfile import ZipFile

from loguru import logger
from sqlalchemy import create_engine
from sqlalchemy.orm import sessionmaker

from . import __db_version__
from .config import Config
from . import exceptions as E
from .models import ModelBase, Pkg, PkgFolder, should_migrate
from .resolvers import Pkg_, CurseResolver, WowiResolver, TukuiResolver, InstawowResolver

if TYPE_CHECKING:
    import aiohttp


_UA_STRING = 'instawow (https://github.com/layday/instawow)'

_loop: cv.ContextVar[asyncio.AbstractEventLoop] = cv.ContextVar('_loop')
_web_client: cv.ContextVar[aiohttp.ClientSession] = cv.ContextVar('_web_client')


def run_in_thread(fn: Callable) -> Callable[..., Awaitable]:
    return lambda *a, **k: _loop.get().run_in_executor(None, partial(fn, *a, **k))


AsyncZipFile = run_in_thread(ZipFile)
AsyncNamedTemporaryFile = partial(run_in_thread(NamedTemporaryFile),
                                  prefix='instawow-')
async_mkdtemp = partial(run_in_thread(mkdtemp), prefix='instawow-')
async_move = run_in_thread(_move)


@asynccontextmanager
async def open_temp_writer() -> AsyncGenerator[Tuple[Path, Callable], None]:
    fh = await AsyncNamedTemporaryFile(delete=False, suffix='.zip')
    try:
        yield (Path(fh.name), run_in_thread(fh.write))
    finally:
        await run_in_thread(fh.close)()


@asynccontextmanager
async def new_temp_dir(delete: bool = False) -> AsyncGenerator[PurePath, None]:
    yield PurePath(await async_mkdtemp())


async def move(paths: Iterable[Path], dest: PurePath) -> None:
    for path in paths:
        logger.debug(f'moving {path} to {dest / path.name}')
        await async_move(str(path), str(dest))


class _Archive:

    def __init__(self, archive: ZipFile, delete_after: bool = False):
        self.archive = archive
        self.delete_after = delete_after
        self.folders = {PurePath(p).parts[0] for p in self.archive.namelist()}

    async def extract(self, parent_folder: Path) -> None:
        def extract() -> None:
            conflicts = self.folders & {f.name for f in parent_folder.iterdir()}
            if conflicts:
                raise E.PkgConflictsWithUncontrolled(conflicts)

            self.archive.extractall(parent_folder)
            self.archive.close()
            if self.delete_after:
                Path(self.archive.filename).unlink()

        return await run_in_thread(extract)()


async def download_archive(url: str, *, chunk_size: int = 4096) -> _Archive:
    if url.startswith('file://'):
        from urllib.parse import unquote

        path = Path(unquote(url[7:]))
        return _Archive(await AsyncZipFile(path))
    else:
        async with _web_client.get().get(url) as response, \
                   open_temp_writer() as (path, write):
            async for chunk in response.content.iter_chunked(chunk_size):
                await write(chunk)

        return _Archive(await AsyncZipFile(path), delete_after=True)


def prepare_db_session(*, config: Config) -> sessionmaker:
    db_url = f"sqlite:///{config.config_dir / 'db.sqlite'}"
    engine = create_engine(db_url)
    ModelBase.metadata.create_all(engine)

    if should_migrate(engine, __db_version__):
        from alembic.autogenerate import compare_metadata
        from alembic.migration import MigrationContext
        from .migrations import make_config, stamp, upgrade

        with engine.begin() as conn:
            alembic_config = make_config(db_url)
            diff = compare_metadata(MigrationContext.configure(conn),
                                    ModelBase.metadata)
            if diff:
                logger.info(f'migrating database to {__db_version__}')
                upgrade(alembic_config, __db_version__)
            else:
                logger.info(f'stamping database with {__db_version__}')
                stamp(alembic_config, __db_version__)

    return sessionmaker(bind=engine)


async def _init_web_client(**kwargs: Any) -> aiohttp.ClientSession:
    from aiohttp import ClientSession, TCPConnector

    kwargs = {'connector': TCPConnector(limit_per_host=10),
              'headers': {'User-Agent': _UA_STRING},
              'trust_env': True,
              **kwargs}
    return ClientSession(**kwargs)


class _ResolverDict(dict):

    def __init__(self, manager: Manager, resolvers: Set[Any]) -> None:
        super().__init__((r.origin, r(manager=manager)) for r in resolvers)

    def __missing__(self, key: Hashable) -> NoReturn:
        raise E.PkgOriginInvalid


class Manager:

    RESOLVERS = {CurseResolver, WowiResolver, TukuiResolver, InstawowResolver}

    def __init__(self, config: Config,
                 web_client_factory: Optional[Callable] = None) -> None:
        self.web_client_factory = web_client_factory or _init_web_client
        self.resolvers = _ResolverDict(self, self.RESOLVERS)

        self.config = config
        Session = prepare_db_session(config=self.config)
        self.db = Session()

    @property
    def web_client(self) -> aiohttp.ClientSession:
        return _web_client.get()

    def run(self, awaitable: Awaitable) -> Any:
        "Run ``awaitable`` inside an explicit context."
        def runner():
            async def arunner():
                async with (await self.web_client_factory()) as client:
                    _web_client.set(client)
                    return await awaitable

            _loop.set(asyncio.get_event_loop())
            return _loop.get().run_until_complete(arunner())

        return cv.copy_context().run(runner)

    def get(self, origin: str, id_or_slug: str) -> Pkg:
        "Retrieve a ``Pkg`` from the database."
        return (self.db.query(Pkg)
                .filter(Pkg.origin == origin,
                        (Pkg.id == id_or_slug) | (Pkg.slug == id_or_slug))
                .first())

    async def resolve(self, origin: str, id_or_slug: str, strategy: str) -> Pkg_:
        "Resolve an ID or slug into a ``Pkg``."
        return await self.resolvers[origin].resolve(id_or_slug, strategy=strategy)

    async def to_install(self, origin: str, id_or_slug: str,
                         strategy: str, replace: bool) -> Callable:
        "Retrieve a package to install."
        async def install() -> E.PkgInstalled:
            pkg = pkg_(folders=[PkgFolder(name=f) for f in sorted(archive.folders)])
            conflicts = (self.db.query(PkgFolder)
                         .filter(PkgFolder.name.in_([f.name for f in pkg.folders]))
                         .first())
            if conflicts:
                raise E.PkgConflictsWithInstalled(conflicts.pkg)

            if replace:
                async with new_temp_dir() as dir_name:
                    await move((self.config.addon_dir / f.name
                                for f in pkg.folders), dir_name)

            await archive.extract(self.config.addon_dir)
            self.db.add(pkg)
            self.db.commit()
            return E.PkgInstalled(pkg)

        if self.get(origin, id_or_slug):
            raise E.PkgAlreadyInstalled
        pkg_ = await self.resolve(origin, id_or_slug, strategy)

        archive = await download_archive(pkg_.download_url)
        return install

    async def to_update(self, origin: str, id_or_slug: str) -> Callable:
        "Retrieve a package to update."
        async def update() -> E.PkgUpdated:
            new_pkg = new_pkg_(folders=[PkgFolder(name=f)
                                        for f in sorted(archive.folders)])
            conflicts = (self.db.query(PkgFolder)
                         .filter(PkgFolder.name.in_([f.name for f in new_pkg.folders]),
                                 PkgFolder.pkg_origin != new_pkg.origin,
                                 PkgFolder.pkg_id != new_pkg.id)
                         .first())
            if conflicts:
                raise E.PkgConflictsWithInstalled(conflicts.pkg)

            async with new_temp_dir() as dir_name:
                await move((self.config.addon_dir / f.name
                            for f in pkg.folders), dir_name)
            await archive.extract(self.config.addon_dir)

            with self.db.begin_nested():
                self.db.delete(pkg)
                self.db.add(new_pkg)

            return E.PkgUpdated(pkg, new_pkg)

        pkg = self.get(origin, id_or_slug)
        if not pkg:
            raise E.PkgNotInstalled

        new_pkg_ = await self.resolve(origin, id_or_slug, pkg.options.strategy)
        if pkg.file_id == new_pkg_.file_id:
            raise E.PkgUpToDate

        archive = await download_archive(new_pkg_.download_url)
        return update

    async def remove(self, origin: str, id_or_slug: str) -> E.PkgRemoved:
        "Remove a package."
        pkg = self.get(origin, id_or_slug)
        if not pkg:
            raise E.PkgNotInstalled

        async with new_temp_dir() as dir_name:
            await move((self.config.addon_dir / f.name
                        for f in pkg.folders), dir_name)

        self.db.delete(pkg)
        self.db.commit()
        return E.PkgRemoved(pkg)


class Bar:

    def __init__(self, *args, **kwargs) -> None:
        kwargs['position'], self.__reset_position = kwargs['position']
        super().__init__(*args, **{'leave': False, 'ascii': True, **kwargs})    # type: ignore

    def close(self) -> None:
        super().close()     # type: ignore
        self.__reset_position()


async def _init_cli_web_client(*, manager: CliManager) -> aiohttp.ClientSession:
    from cgi import parse_header
    from aiohttp import TraceConfig

    async def do_on_request_end(_session, _ctx, params: aiohttp.TraceRequestEndParams) -> None:
        if params.response.content_type in {
                'application/zip',
                # Curse at it again
                'application/x-amz-json-1.0'}:
            cd = params.response.headers.get('Content-Disposition', '')
            _, cd_params = parse_header(cd)
            filename = cd_params.get('filename') or params.response.url.name

            bar = manager.Bar(total=params.response.content_length,
                              desc=f'  Downloading {filename}',
                              miniters=1, unit='B', unit_scale=True,
                              position=manager._get_bar_position())

            async def ticker(bar=bar, params=params) -> None:
                while not params.response.content._eof:
                    bar.update(params.response.content._cursor - bar.n)
                    await asyncio.sleep(bar.mininterval)
                bar.close()
            _loop.get().create_task(ticker())

    trace_config = TraceConfig()
    trace_config.on_request_end.append(do_on_request_end)
    trace_config.freeze()
    return await _init_web_client(trace_configs=[trace_config])


class SafeFuture(asyncio.Future):

    def result(self) -> object:
        return self.exception() or super().result()

    async def intercept(self, awaitable: Awaitable) -> SafeFuture:
        try:
            self.set_result(await awaitable)
        except E.ManagerError as error:
            self.set_exception(error)
        except Exception as error:
            logger.exception('internal error')
            self.set_exception(E.InternalError(error=error))
        return self


class CliManager(Manager):

    def __init__(self, config: Config, show_progress: bool = True) -> None:
        super().__init__(config=config,
                         web_client_factory=(partial(_init_cli_web_client, manager=self)
                                             if show_progress else None))
        self.show_progress = show_progress
        self._bar_positions = [False]

        from tqdm import tqdm
        self.Bar = type('Bar', (Bar, tqdm), {})

    def _get_bar_position(self) -> Tuple[int, Callable]:
        # Get the first available bar slot
        try:
            b = self._bar_positions.index(False)
            self._bar_positions[b] = True
        except ValueError:
            b = len(self._bar_positions)
            self._bar_positions.append(True)
        return (b, lambda b=b: self._bar_positions.__setitem__(b, False))

    async def gather(self, it: Iterable, *,
                     desc: Optional[str] = None) -> List[SafeFuture]:
        async def intercept(coro, index, bar):
            future = await SafeFuture().intercept(coro)
            bar.update(1)
            return index, future

        coros = list(it)
        with self.Bar(total=len(coros), disable=not self.show_progress,
                      position=self._get_bar_position(), desc=desc) as bar:
            futures = [intercept(c, i, bar) for i, c in enumerate(coros)]
            results = [v for _, v in sorted([await r for r in
                                             asyncio.as_completed(futures)])]

            # Wait for ``ticker``s to complete so all bars get to wipe
            # their pretty faces off the face of the screen
            while len(asyncio.all_tasks()) > 1:
                await asyncio.sleep(bar.mininterval)
        return results

    def resolve_many(self, values: Iterable) -> List[Union[E.ManagerResult, Pkg]]:
        async def resolve_many():
            return [r.result() for r in
                    (await self.gather((self.resolve(*a) for a in values),
                                       desc='Resolving'))]

        return self.run(resolve_many())

    def install_many(self, values: Iterable) -> List[E.ManagerResult]:
        async def install_many():
            return [(r if r.exception() else
                     await SafeFuture().intercept(r.result()())
                     ).result()
                    for r in (await self.gather((self.to_install(*a) for a in values),
                                                desc='Fetching'))]

        return self.run(install_many())

    def update_many(self, values: Iterable) -> List[E.ManagerResult]:
        async def update_many():
            return [(r if r.exception() else
                     await SafeFuture().intercept(r.result()())
                     ).result()
                    for r in (await self.gather((self.to_update(*a) for a in values),
                                                desc='Checking'))]

        return self.run(update_many())


class WsManager(Manager):

    def __init__(self, loop: Optional[asyncio.AbstractEventLoop] = None) -> None:
        self.web_client_factory = _init_web_client
        self.resolvers = _ResolverDict(self, self.RESOLVERS)

    def finalise(self, config: Config) -> WsManager:
        self.config = config
        Session = prepare_db_session(config=self.config)
        self.db = Session()
        return self

    async def _poll(self, web_request: aiohttp.web.Request) -> None:
        import aiohttp
        from .api import (ErrorCodes, ApiError,
                          Request, InstallRequest, UpdateRequest, RemoveRequest,
                          SuccessResponse, ErrorResponse,
                          jsonify, parse_request)

        TR = TypeVar('TR', bound=Request)

        async def respond(request: TR, awaitable: Awaitable) -> None:
            try:
                result = await awaitable
            except ApiError as error:
                response = ErrorResponse.from_api_error(error)
            except E.ManagerError as error:
                values = {'id': request.id,
                          'error': {'code': ErrorCodes[type(error).__name__],
                                    'message': error.message}}
                response = ErrorResponse(**values)
            except Exception:
                logger.exception('internal error')
                values = {'id': request.id,
                          'error': {'code': ErrorCodes.INTERNAL_ERROR,
                                    'message': 'encountered an internal error'}}
                response = ErrorResponse(**values)
            else:
                response = request.consume_result(result)
            await websocket.send_json(response, dumps=jsonify)

        async def receiver() -> None:
            async for message in websocket:
                if message.type == aiohttp.WSMsgType.TEXT:
                    try:
                        request = parse_request(message.data)
                    except ApiError as error:
                        async def schedule_error(error=error):
                            raise error
                        _loop.get().create_task(respond(None, schedule_error()))  # type: ignore
                    else:
                        if request.__class__ in {InstallRequest, UpdateRequest, RemoveRequest}:
                            # Here we're wrapping the coroutine in a future
                            # that the consumer will `wait_for` to complete therefore
                            # preserving the order of (would-be) synchronous operations
                            future = _loop.get().create_future()
                            consumer_queue.put_nowait((request, future))

                            async def schedule(request=request, future=future):
                                try:
                                    future.set_result(await request.prepare_response(self))
                                except Exception as error:
                                    future.set_exception(error)
                            _loop.get().create_task(schedule())
                        else:
                            _loop.get().create_task(respond(request,
                                                            request.prepare_response(self)))

        async def consumer() -> None:
            while True:
                request, future = await consumer_queue.get()

                async def consume(request=request, future=future):
                    result = await asyncio.wait_for(future, None)
                    if request.__class__ in {InstallRequest, UpdateRequest}:
                        result = await result()
                    return result
                await respond(request, consume())
                consumer_queue.task_done()

        consumer_queue: asyncio.Queue[Tuple[Request,
                                            asyncio.Future]] = asyncio.Queue()
        websocket = aiohttp.web.WebSocketResponse()

        await websocket.prepare(web_request)
        _loop.get().create_task(consumer())
        await receiver()

    def serve(self, host: str = '127.0.0.1', port: Optional[int] = None) -> None:
        async def aserve():
            import os
            import socket
            from aiohttp import web

            app = web.Application()
            app.router.add_routes([web.get('/', self._poll)])    # type: ignore
            app_runner = web.AppRunner(app)
            await app_runner.setup()

            server = await _loop.get().create_server(app_runner.server, host, port,
                                                     family=socket.AF_INET)
            sock = server.sockets[0]
            message = ('{{"address": "ws://{}:{}/"}}\n'
                       .format(*sock.getsockname()).encode())
            try:
                # Try sending message over fd 3 for IPC with Node
                # and if that fails...
                os.write(3, message)
            except OSError:
                # ... write to stdout
                os.write(1, message)

            try:
                await server.serve_forever()
            except (KeyboardInterrupt, SystemExit):
                pass
            finally:
                await app_runner.cleanup()

        self.run(aserve())




from alembic import context
from sqlalchemy import engine_from_config, pool

from instawow.models import ModelBase


target_metadata = ModelBase.metadata


def run_offline() -> None:
    url = context.config.get_main_option('sqlalchemy.url')
    context.configure(url=url, target_metadata=target_metadata, literal_binds=True)
    with context.begin_transaction():
        context.run_migrations()


def run_online() -> None:
    alembic_config = context.config.get_section(context.config.config_ini_section)
    engine = engine_from_config(alembic_config,
                                prefix='sqlalchemy.',
                                poolclass=pool.NullPool)
    with engine.connect() as connection:
        context.configure(connection=connection,
                          target_metadata=target_metadata)
        with context.begin_transaction():
            context.run_migrations()


def main() -> None:
    if context.is_offline_mode():
        run_offline()
    else:
        run_online()


if __name__ == 'env_py':    # Alembic appears to mangle the module's `__name__`
    main()




from pathlib import Path
from setuptools import find_packages, setup

try:
    import fastentrypoints
except ImportError:
    pass

from instawow import __version__


setup(name='instawow',
      version=__version__,
      description='A CLI for managing World of Warcraft add-ons.',
      url='http://github.com/layday/instawow',
      author='layday',
      author_email='layday@protonmail.com',
      license='GPL-3.0-or-later',
      long_description=Path('README.rst').read_text(),
      python_requires='~=3.7',
      packages=find_packages(),
      install_requires=Path('requirements.txt').read_text(),
      include_package_data=True,
      entry_points={'console_scripts': ['instawow = instawow.cli:main']})
